"""Re-exports a subset of the functionality of the following packages.

- beartype
- jaxtyping

We additionally rename some of the functions to be independent of the underlying packages. For example,
it should be easily possible to switch from beartype to typeguard for runtime type checking.
"""

from warnings import filterwarnings

# re-export everything necessary from beartype, never use beartype itself.
# additionally ignore the default beartype deprecation warnings.
from beartype.door import (
    die_if_unbearable as assert_instance,
    is_bearable as is_instance,
)
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from beartype.vale import (
    Is,
    IsAttr,
    IsEqual,
    IsInstance,
    IsSubclass,
)

# re-export everything necessary from jaxtyping, never use jaxtyping itself
from jaxtyping import (
    AbstractDtype,  # type: ignore[reportGeneralTypeIssues]
    Bool,  # type: ignore[reportGeneralTypeIssues]
    Complex,  # type: ignore[reportGeneralTypeIssues]
    Complex64,  # type: ignore[reportGeneralTypeIssues]
    Complex128,  # type: ignore[reportGeneralTypeIssues]
    Float,  # type: ignore[reportGeneralTypeIssues]
    Float16,  # type: ignore[reportGeneralTypeIssues]
    Float32,  # type: ignore[reportGeneralTypeIssues]
    Float64,  # type: ignore[reportGeneralTypeIssues]
    Inexact,  # type: ignore[reportGeneralTypeIssues]
    Int,  # type: ignore[reportGeneralTypeIssues]
    Int8,  # type: ignore[reportGeneralTypeIssues]
    Int16,  # type: ignore[reportGeneralTypeIssues]
    Int32,  # type: ignore[reportGeneralTypeIssues]
    Int64,  # type: ignore[reportGeneralTypeIssues]
    Integer,  # type: ignore[reportGeneralTypeIssues]
    Num,  # type: ignore[reportGeneralTypeIssues]
    Real,  # type: ignore[reportGeneralTypeIssues]
    Shaped,  # type: ignore[reportGeneralTypeIssues]
    UInt,  # type: ignore[reportGeneralTypeIssues]
    UInt8,  # type: ignore[reportGeneralTypeIssues]
    UInt16,  # type: ignore[reportGeneralTypeIssues]
    UInt32,  # type: ignore[reportGeneralTypeIssues]
    UInt64,  # type: ignore[reportGeneralTypeIssues]
)

from ._overload import typecheck_overload
from ._typecheck import typecheck

filterwarnings("ignore", category=BeartypeDecorHintPep585DeprecationWarning)

__all__ = [
    # decorators (runtime type-checking)
    "typecheck",
    "typecheck_overload",
    # introspection
    "is_instance",  # like "isinstance(...)"
    "assert_instance",  # like "assert isinstance(...)"
    # validators (runtime only, see https://beartype.readthedocs.io/en/latest/api_vale/)
    "Is",  # Annotated[str, Is[lambda x: x > 0)]]
    "IsAttr",  # Annotated[NumpyArray, IsAttr["ndim", IsEqual[1]]]
    "IsEqual",  # Annotated[list, IsEqual[list(range(42))]]
    "IsSubclass",  # Annotated[type, IsSubclass[str, bytes]]
    "IsInstance",  # Annotated[object, IsInstance[str, bytes]]
    # abstract type
    "AbstractDtype",
    # union array types
    "Shaped",  # Any type at all (e.g. object or string)
    "Num",  # Any integer, unsigned integer, floating, or complex
    "Real",  # Any integer, unsigned integer or floating
    "Inexact",  # Any floating or complex
    "Float",  # Any floating point
    "Complex",  # Any complex
    "Integer",  # Any integer or unsigned integer
    "UInt",  # Any unsigned integer
    "Int",  # Any signed integer
    # concrete array types
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "Float16",
    "Float32",
    "Float64",
    "Bool",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
    "Complex64",
    "Complex128",
]

try:
    from numpy import ndarray as NumpyArray  # noqa: N812

    __all__ += ["NumpyArray"]
except ImportError:  # pragma: no cover
    ...

try:
    from torch import Tensor as TorchArray

    __all__ += ["TorchArray"]
except ImportError:  # pragma: no cover
    ...

try:
    from jax import Array as JaxArray

    __all__ += ["JaxArray"]
except ImportError:  # pragma: no cover
    ...


def get_version() -> str:
    """Return the package version or "unknown" if no version can be found."""
    from importlib import metadata

    try:
        return metadata.version(__name__)
    except metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


__version__: str = get_version()



"""Enable typechecking of @overload-decorated functions.

Previously, this library supported [plum](https://github.com/beartype/plum) as a multiple dispatch library, but it
turns out that it's really difficult to implement multiple dispatch in a generic way, see:
https://discuss.python.org/t/multiple-dispatch-based-on-typing-overload/26197/17.

Even overloads are tricky to determine and very slightly between type checking libraries, see for example:
https://microsoft.github.io/pyright/#/type-concepts-advanced?id=overloads

For the stated reasons, we choose to implement a very simple form of typechecking for overloaded functions that
works as follows:

1. The existing overloads are determined using ``get_overloads``
2. The overloads are iterated in definition-order until a matching overload is found.
3. If a matching overload is found (typecheck successful) return the result, otherwise raise an error.
"""

import sys
import warnings
from functools import wraps
from inspect import signature

from beartype._util.func.arg.utilfuncargiter import ArgKind, iter_func_args
from beartype.roar import BeartypeCallHintParamViolation, BeartypeCallHintReturnViolation
from beartype.typing import Any, Iterable, Sequence, Tuple
from jaxtyping import TypeCheckError  # type: ignore[reportGeneralTypeIssues]
from typing_extensions import Never, get_overloads

from ._typecheck import CallableAnyT, raise_if_missing_annotation, typecheck

__all__ = [
    "typecheck_overload",
]


class UnavailableOverloadError(Exception): ...


class MissingOverloadError(Exception): ...


class IncompatibleOverloadError(Exception): ...


def typecheck_overload(fn: CallableAnyT) -> CallableAnyT:
    """Ensure that an implementing function satisfied at least one of its defined overloads.

    To check if the implementing function satisfies one of the overloads, we iterate over all overloads (in definition
    order) and check if the signature matches the provided ``*args`` and ``**kwargs``. If the signature matches,
    the implementing function ``fn`` is called and the return value is cached because we also type-check the return
    value and the first return value might not satisfy the return type for the current overload. If further signatures
    match an already computed ``fn``, we directly return the cached return value and check if the return type matches
    the current overload.

    :param fn: An implementing function for @overload-decorated specifications.
    :return: Typechecked function or method.
    :raises: MissingAnnotationError if ``fn`` does not contain ``__annotations__`` (is no function or method).
    :raises: MissingOverloadError if no overload can be determined for ``fn``.
    :raises: IncompatibleOverloadError if one of the @overload functions is incompatible with the signature of ``fn``.
    :raises: UnavailableOverloadError if no overload matches the input arguments to ``fn``.
    """
    # this check must be before the ``len(fn.__annotations__)``, to ensure ``fn`` has ``__annotations__``
    raise_if_missing_annotation(fn)

    # make sure that the implementing fn is not annotated
    if len(fn.__annotations__) > 0:
        warn_if_unexpected_annotation(fn)

    # we are not yet sure if we are dealing with a function, therefore we wrap ``get_overloads`` safely
    overloads = safe_get_overloads(fn)

    # make sure that at least one overload is found (registered)
    if len(overloads) == 0:
        raise_if_no_overload(fn)

    # make sure that overloads contain equal parameters as in the implementation
    raise_if_incompatible_overload(fn, overloads)

    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        result = Never

        def cached_result(*_: Any, **__: Any) -> Any:
            # we cache the result of fn(*args, **kwargs), such that it only has to be computed once, even though
            # we might need the result value for multiple return-type checks
            nonlocal result
            result = fn(*args, **kwargs) if result is Never else result
            return result

        for f in overloads:
            # wrapping the overload function preserves the original signature, and we can add an implementation
            # that returns a cached result of the non-typechecked implementing function ``fn` using a closure
            cache: CallableAnyT = wraps(f)(cached_result)
            try:
                return typecheck(cache, skip_annotation_check=True)(*args, **kwargs)
            except (BeartypeCallHintParamViolation, BeartypeCallHintReturnViolation, TypeCheckError):
                ...

        arg_msg = "without arguments"
        if (has_args := len(args) > 0) ^ (len(kwargs) > 0):
            arg_msg = f"with arguments '{args if has_args else kwargs}'"

        if len(args) > 0 and len(kwargs) > 0:
            arg_msg = f"with arguments '{args}' and keywords '{kwargs}'"

        if result is not Never:
            arg_msg += f"and return value '{result}'"

        msg = (
            f"No suitable overload was found for @typecheck_overload-decorated function '{fn.__qualname__}' {arg_msg}, "
            f"the available overloads are:\n" + "\n".join(str(signature(o)) for o in overloads)
        )
        raise UnavailableOverloadError(msg)

    return wrapper


def warn_if_unexpected_annotation(fn: CallableAnyT) -> None:
    msg = (
        f"The function implementing @overload-decorated definitions should not contain type annotations, "
        f"because the behaviour is not defined, but found annotations '{fn.__annotations__}' which are ignored. "
    )
    warnings.warn(msg, stacklevel=2)


def raise_if_no_overload(fn: CallableAnyT) -> None:
    if sys.version_info < (3, 11):
        additional_help = (
            "Did you use 'from typing import overload'? If this is the case, use 'from typing_extensions import "
            "overload' instead. Using @overload from typing before Python 3.11 does not make the overload visible "
            "to safecheck. When you upgrade to Python 3.11, you'll be able to use 'from typing import overload'."
        )
    else:
        additional_help = "Did you forget to use '@overload'?"

    msg = f"Could not find any overload for '{fn.__qualname__}'." + additional_help
    raise MissingOverloadError(msg)


def raise_if_incompatible_overload(fn: CallableAnyT, overloads: Iterable[CallableAnyT]) -> None:
    args_fn = tuple(iter_func_args_no_default_value(fn))
    for fn_overload in overloads:
        args_ov = tuple(iter_func_args_no_default_value(fn_overload))
        if args_ov != args_fn:
            signature_fn = signature(fn)
            signature_overload = signature(fn_overload)
            msg = (
                "Signature of this @overload-decorated function is not compatible with the implementation, found an "
                f"implementation with signature '{signature_fn}' and an incompatible overload with signature "
                f"'{signature_overload}', make sure that the @overload-decorated functions share the signature "
                f"with the implementing function."
            )
            raise IncompatibleOverloadError(msg)


def iter_func_args_no_default_value(fn: CallableAnyT) -> Iterable[Tuple[ArgKind, str]]:
    # it should be safe to use is_unwrap=False, because we don't use the default value
    for arg in iter_func_args(fn, is_unwrap=False):
        # [:2] extracts the argument-kind and -name from the resulting tuple
        yield arg[:2]


def safe_get_overloads(fn: CallableAnyT) -> Sequence[CallableAnyT]:
    try:
        return get_overloads(fn)
    except AttributeError:
        msg = (
            f"Could not determine overloads for '{fn}' because of missing attributes, maybe the object is no "
            f"function or method?"
        )
    raise MissingOverloadError(msg)



from beartype import beartype as _typecheck
from beartype.typing import Any, Callable, TypeVar, Union
from jaxtyping import jaxtyped as _shapecheck

__all__ = [
    "typecheck",
    "CallableAnyT",
    "CheckableT",
]


class MissingAnnotationError(Exception): ...


CallableAnyT = Callable[..., Any]
CheckableT = TypeVar(
    "CheckableT",
    # Equal to ``BeartypeableT``, but without allowing class decorators
    bound=Union[
        # An arbitrary callable *OR*...
        CallableAnyT,
        # A C-based unbound class method descriptor (i.e., a pure-Python unbound
        # function decorated by the builtin @classmethod decorator) *OR*...
        classmethod,
        # A C-based unbound property method descriptor (i.e., a pure-Python
        # unbound function decorated by the builtin @property decorator) *OR*...
        property,
        # A C-based unbound static method descriptor (i.e., a pure-Python
        # unbound function decorated by the builtin @staticmethod decorator).
        staticmethod,
    ],
)


def raise_if_missing_annotation(fn: CheckableT) -> None:  # type: ignore[reportInvalidTypeVarUse]
    """Any function or method should have an ``__annotations__`` attribute, we therefore assume it's a function."""
    if not hasattr(fn, "__annotations__"):
        msg = (
            f"@typecheck-decorated objects must have an '__annotations__' attribute, but '{fn}' has "
            f" no '__annotations__' attribute. Did you try to use @typecheck on an unsupported object? "
            f"In contrast to beartype, we do not allow implicit typechecking of entire classes and @typecheck "
            f"must not be used as a class decorator."
        )
        raise MissingAnnotationError(msg)


def typecheck(fn: CheckableT, *, skip_annotation_check: bool = False) -> CheckableT:
    """Typecheck a function without jaxtyping annotations, otherwise additionally shapecheck the function.

    :param fn: Any function or method.
    :param skip_annotation_check: Disable the missing annotation check if you already know it exists.
    :return: Typechecked function or method.
    :raises: MissingAnnotationError if ``fn`` does not contain ``__annotations__`` (is no function or method).
    :raises: BeartypeException if a call to the function does not satisfy the typecheck.
    """
    # check if there is any annotation requiring a shapecheck, i.e. any jaxtyping annotation that is not "..."
    # this check is significantly slower than the string-based check implemented below (~+50%), but this should
    # only be relevant in tight loops.
    # for annotation in fn.__annotations__.values():
    #     if getattr(annotation, "dim_str", "") != "...":

    if not skip_annotation_check:
        raise_if_missing_annotation(fn)

    # simply check if there is any mention of jaxtyping in the annotations, this adds barely any overhead to
    # a base call of beartype's @beartype
    if "jaxtyping" in str(fn.__annotations__):
        # shapecheck implies typecheck
        return _shapecheck(typechecker=_typecheck)(fn)  # type: ignore[reportGeneralTypeIssues]

    return _typecheck(fn)
